"""
---
title: Gated Linear Units and Variants
summary: >
  Train an auto-regressive transformer with Gated Linear Units and variants
  for the position-wise feedforward network (FFN).
---

# Gated Linear Units and Variants

* [Experiment that uses `labml.configs`](experiment.html)
* [Simpler version from scratch](simple.html)
"""
