"""
---
title: Scripts to show example usages stable diffusion
summary: >
 Annotated PyTorch implementation/tutorial of example usages of stable diffusion
---

# Scripts to show example usages [stable diffusion](../index.html)

* [Prompt to image diffusion](text_to_image.html)
* [Image to image diffusion](image_to_image.html)
* [In-painting](in_paint.html)
"""
