"""
---
title: Neural Networks with Uncertainty Estimation
summary: >
 A set of PyTorch implementations/tutorials related to uncertainty estimation
---

# Neural Networks with Uncertainty Estimation

These are neural network architectures that estimate the uncertainty of the predictions.

* [Evidential Deep Learning to Quantify Classification Uncertainty](evidence/index.html)
"""
