"""
---
title: Generative Adversarial Networks
summary: >
 A set of PyTorch implementations/tutorials of GANs.
---

# Generative Adversarial Networks

* [Original GAN](original/index.html)
* [GAN with deep convolutional network](dcgan/index.html)
* [Cycle GAN](cycle_gan/index.html)
* [Wasserstein GAN](wasserstein/index.html)
* [Wasserstein GAN with Gradient Penalty](wasserstein/gradient_penalty/index.html)
* [StyleGAN 2](stylegan/index.html)
"""