"""
---
title: Samples
summary: >
    Samples for inference and fine-tuning
---

# Samples

* [Generating text](generate.html)
* [Fine tuning the biases with pipeline-parallel training](finetune.html)
"""