"""
---
title: Neural Networks with Adaptive Computation
summary: >
 A set of PyTorch implementations/tutorials related to adaptive computation
---

# Neural Networks with Adaptive Computation

These are neural network architectures that change the computation complexity based on the
complexity of the input sample.

* 🚧 TODO: Adaptive Computation Time for Recurrent Neural Networks
* [PonderNet: Learning to Ponder](ponder_net/index.html)
"""
