"""
---
title: Modules used in stable diffusion
summary: >
 Models and components for stable diffusion.
---

# [Stable Diffusion](../index.html) Models

* [AutoEncoder](autoencoder.html)
* [U-Net](unet.html) with [attention](unet_attention.html)
* [CLIP embedder](clip_embedder.html).
"""
